@@ -197,6 +197,19 @@ class Record(RecordWithoutPickling):
     def __ne__(self, other):
         return not self.__eq__(other)
 
+
+class ImmutableRecordWithoutPickling(RecordWithoutPickling):
+    "Hashable record. Does not explicitly enforce immutability."
+
+    def __hash__(self):
+        return hash(
+                (type(self),) + tuple(getattr(self, field)
+                    for field in self.__class__.fields))
+
+
+class ImmutableRecord(ImmutableRecordWithoutPickling, Record):
+    pass
+
 # }}}
 
 
@@ -441,6 +454,7 @@ def memoize(*args, **kwargs):
     raise TypeError(
         "memoize received unexpected position arguments: %s" % args)
 
+
 FunctionValueCache = memoize
 
 
@@ -448,44 +462,7 @@ class _HasKwargs(object):
     pass
 
 
-def memoize_method(method):
-    """Supports cache deletion via ``method_name.clear_cache(self)``.
-
-    .. note::
-        *clear_cache* support requires Python 2.5 or newer.
-    """
-
-    cache_dict_name = intern("_memoize_dic_"+method.__name__)
-
-    def wrapper(self, *args, **kwargs):
-        if kwargs:
-            key = (_HasKwargs, frozenset(six.iteritems(kwargs))) + args
-        else:
-            key = args
-
-        try:
-            return getattr(self, cache_dict_name)[key]
-        except AttributeError:
-            result = method(self, *args, **kwargs)
-            setattr(self, cache_dict_name, {key: result})
-            return result
-        except KeyError:
-            result = method(self, *args, **kwargs)
-            getattr(self, cache_dict_name)[key] = result
-            return result
-
-    def clear_cache(self):
-        delattr(self, cache_dict_name)
-
-    if sys.version_info >= (2, 5):
-        from functools import update_wrapper
-        new_wrapper = update_wrapper(wrapper, method)
-        new_wrapper.clear_cache = clear_cache
-
-    return new_wrapper
-
-
-def memoize_on_first_arg(function):
+def memoize_on_first_arg(function, cache_dict_name=None):
     """Like :func:`memoize_method`, but for functions that take the object
     to do memoization as first argument.
 
@@ -495,8 +472,9 @@ def memoize_on_first_arg(function):
         *clear_cache* support requires Python 2.5 or newer.
     """
 
-    cache_dict_name = intern("_memoize_dic_"
-            + function.__module__ + function.__name__)
+    if cache_dict_name is None:
+        cache_dict_name = intern("_memoize_dic_"
+                + function.__module__ + function.__name__)
 
     def wrapper(obj, *args, **kwargs):
         if kwargs:
@@ -526,6 +504,16 @@ def memoize_on_first_arg(function):
     return new_wrapper
 
 
+def memoize_method(method):
+    """Supports cache deletion via ``method_name.clear_cache(self)``.
+
+    .. note::
+        *clear_cache* support requires Python 2.5 or newer.
+    """
+
+    return memoize_on_first_arg(method, intern("_memoize_dic_"+method.__name__))
+
+
 def memoize_method_with_uncached(uncached_args=[], uncached_kwargs=set()):
     """Supports cache deletion via ``method_name.clear_cache(self)``.
 
@@ -627,8 +615,7 @@ def memoize_method_nested(inner):
 
 class memoize_in(object):  # noqa
     """Adds a cache to a function nested inside a method. The cache is attached
-    to *memoize_cache_context* (if it exists) or *self* in the outer (method)
-    namespace.
+    to *object*.
 
     Requires Python 2.5 or newer.
     """
@@ -804,21 +791,8 @@ def product(iterable):
     return reduce(mul, iterable, 1)
 
 
-try:
-    all = __builtins__.all
-    any = __builtins__.any
-except AttributeError:
-    def all(iterable):
-        for i in iterable:
-            if not i:
-                return False
-        return True
-
-    def any(iterable):
-        for i in iterable:
-            if i:
-                return True
-        return False
+all = six.moves.builtins.all
+any = six.moves.builtins.any
 
 
 def reverse_dictionary(the_dict):
@@ -851,7 +825,7 @@ def uniform_interval_splitting(n, granularity, max_intervals):
 
     and *interval_size* is a multiple of *granularity*.
     """
-    # ported from Thrust -- minor Apache v2 license violation
+    # ported from Thrust
 
     grains = div_ceil(n, granularity)
 
@@ -1072,6 +1046,9 @@ def wandering_element(length, wanderer=1, landscape=0):
 
 
 def indices_in_shape(shape):
+    if isinstance(shape, int):
+        shape = (shape,)
+
     if len(shape) == 0:
         yield ()
     elif len(shape) == 1:
@@ -1157,6 +1134,7 @@ def generate_all_nonnegative_integer_tuples(length, least=0):
                     yield prebase + [current_max] + postbase
         current_max += 1
 
+
 # backwards compatibility
 generate_positive_integer_tuples_below = generate_nonnegative_integer_tuples_below
 generate_all_positive_integer_tuples = generate_all_nonnegative_integer_tuples
@@ -1403,9 +1381,9 @@ def string_histogram(iterable, min_value=None, max_value=None,
             full = int(floor(scaled))
             eighths = int(ceil((scaled-full)*8))
             if eighths:
-                return full*unichr(0x2588) + unichr(0x2588+(8-eighths))
+                return full*six.unichr(0x2588) + six.unichr(0x2588+(8-eighths))
             else:
-                return full*unichr(0x2588)
+                return full*six.unichr(0x2588)
     else:
         def format_bar(cnt):
             return int(ceil(cnt*width/max_count))*"#"
@@ -1771,6 +1749,7 @@ def generate_unique_names(prefix):
         yield "%s_%d" % (prefix, try_num)
         try_num += 1
 
+
 generate_unique_possibilities = MovedFunctionDeprecationWrapper(
         generate_unique_names)
 
@@ -1814,6 +1793,7 @@ def _test():
     import doctest
     doctest.testmod()
 
+
 if __name__ == "__main__":
     _test()
 
