@@ -1,6 +1,8 @@
 """Generic persistent, concurrent dictionary-like facility."""
 
 from __future__ import division, with_statement
+from __future__ import absolute_import
+import six
 
 __copyright__ = "Copyright (C) 2011,2014 Andreas Kloeckner"
 
@@ -30,6 +32,7 @@ logger = logging.getLogger(__name__)
 
 import sys
 import os
+import errno
 
 try:
     import hashlib
@@ -54,7 +57,7 @@ def _erase_dir(dir):
 
 
 def update_checksum(checksum, obj):
-    if isinstance(obj, unicode):
+    if isinstance(obj, six.text_type):
         checksum.update(obj.encode("utf8"))
     else:
         checksum.update(obj)
@@ -105,6 +108,10 @@ class LockManager(CleanupBase):
                     from warnings import warn
                     warn("could not obtain lock--delete '%s' if necessary"
                             % self.lock_file)
+                if attempts > 3 * 60:
+                    raise RuntimeError("waited more than three minutes "
+                            "on the lock file '%s'"
+                            "--something is wrong" % self.lock_file)
 
             cleanup_m.register(self)
 
@@ -120,23 +127,35 @@ class LockManager(CleanupBase):
 class ItemDirManager(CleanupBase):
     def __init__(self, cleanup_m, path):
         from os import mkdir
+        import errno
 
         self.path = path
         try:
             mkdir(self.path)
+        except OSError as e:
+            if e.errno != errno.EEXIST:
+                raise
+            self.existed = True
+        else:
             cleanup_m.register(self)
             self.existed = False
-        except OSError:
-            self.existed = True
 
     def sub(self, n):
         from os.path import join
         return join(self.path, n)
 
     def reset(self):
-        import os
-        _erase_dir(self.path)
-        os.mkdir(self.path)
+        try:
+            _erase_dir(self.path)
+        except OSError as e:
+            if e.errno != errno.ENOENT:
+                raise
+
+        try:
+            os.mkdir(self.path)
+        except OSError as e:
+            if e.errno != errno.EEXIST:
+                raise
 
     def clean_up(self):
         pass
@@ -189,6 +208,9 @@ class KeyBuilder(object):
     if sys.version_info >= (3,):
         def update_for_str(self, key_hash, key):
             key_hash.update(key.encode('utf8'))
+
+        def update_for_bytes(self, key_hash, key):
+            key_hash.update(key)
     else:
         def update_for_str(self, key_hash, key):
             key_hash.update(key)
@@ -204,11 +226,11 @@ class KeyBuilder(object):
         for set_key in sorted(key):
             self.rec(key_hash, set_key)
 
-    def update_for_NoneType(self, key_hash, key):
+    def update_for_NoneType(self, key_hash, key):  # noqa
         key_hash.update("<None>".encode('utf8'))
 
     def update_for_dtype(self, key_hash, key):
-        return key.str.encode("utf8")
+        key_hash.update(key.str.encode('utf8'))
 
     # }}}
 
@@ -254,7 +276,7 @@ class PersistentDict(object):
 
         try:
             os.makedirs(self.container_dir)
-        except OSError, e:
+        except OSError as e:
             from errno import EEXIST
             if e.errno != EEXIST:
                 raise
@@ -276,17 +298,20 @@ class PersistentDict(object):
                 if item_dir_m.existed:
                     item_dir_m.reset()
 
-                for info_name, info_value in info_files.iteritems():
+                for info_name, info_value in six.iteritems(info_files):
                     info_path = item_dir_m.sub("info_"+info_name)
 
                     with open(info_path, "wt") as outf:
                         outf.write(info_value)
 
-                from cPickle import dump, HIGHEST_PROTOCOL
+                from six.moves.cPickle import dump, HIGHEST_PROTOCOL
                 value_path = item_dir_m.sub("contents")
                 with open(value_path, "wb") as outf:
                     dump(value, outf, protocol=HIGHEST_PROTOCOL)
 
+                logger.debug("%s: cache store [key=%s]" % (
+                    self.identifier, hexdigest_key))
+
                 # Write key last, so that if the reader below
                 key_path = item_dir_m.sub("key")
                 with open(key_path, "wb") as outf:
@@ -304,6 +329,8 @@ class PersistentDict(object):
         from os.path import join, isdir
         item_dir = join(self.container_dir, hexdigest_key)
         if not isdir(item_dir):
+            logger.debug("%s: cache miss [key=%s]" % (
+                self.identifier, hexdigest_key))
             raise NoSuchEntryError(key)
 
         cleanup_m = CleanupManager()
@@ -315,7 +342,7 @@ class PersistentDict(object):
                 key_path = item_dir_m.sub("key")
                 value_path = item_dir_m.sub("contents")
 
-                from cPickle import load
+                from six.moves.cPickle import load
 
                 # {{{ load key file
 
@@ -324,9 +351,9 @@ class PersistentDict(object):
                 try:
                     with open(key_path, "rb") as inf:
                         read_key = load(inf)
-                except IOError, e:
+                except IOError as e:
                     exc = e
-                except EOFError, e:
+                except EOFError as e:
                     exc = e
 
                 if exc is not None:
@@ -342,10 +369,19 @@ class PersistentDict(object):
 
                 if read_key != key:
                     # Key collision, oh well.
-                    logger.debug("key collsion in cache at '%s'"
-                            % self.container_dir)
+                    from warnings import warn
+                    warn("%s: key collision in cache at '%s' -- these are "
+                            "sufficiently unlikely that they're often "
+                            "indicative of a broken implementation "
+                            "of equality comparison"
+                            % (self.identifier, self.container_dir))
+                    # This is here so we can debug the equality comparison
+                    read_key == key
                     raise NoSuchEntryError(key)
 
+                logger.debug("%s: cache hit [key=%s]" % (
+                    self.identifier, hexdigest_key))
+
                 # {{{ load value
 
                 exc = None
@@ -353,9 +389,9 @@ class PersistentDict(object):
                 try:
                     with open(value_path, "rb") as inf:
                         read_contents = load(inf)
-                except IOError, e:
+                except IOError as e:
                     exc = e
-                except EOFError, e:
+                except EOFError as e:
                     exc = e
 
                 if exc is not None:
@@ -409,7 +445,12 @@ class PersistentDict(object):
         self.remove(key)
 
     def clear(self):
-        _erase_dir(self.container_dir)
+        try:
+            _erase_dir(self.container_dir)
+        except OSError as e:
+            if e.errno != errno.ENOENT:
+                raise
+
         self._make_container_dir()
 
 # }}}
